# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to morphologically analyze surface forms of Turkish words."""

from typing import List, Optional

from turkish_morphology import fst


def _remove_proper_feature(human_readable: str) -> str:
  """Removes proper feature from human-readable analysis."""
  human_readable = human_readable.replace("+[Proper=False]", "")
  human_readable = human_readable.replace("+[Proper=True]", "")
  return human_readable


def surface_form(surface_form: str,
                 use_proper_feature: Optional[bool] = True) -> List[str]:
  """Morphologically analyses given surface form.

  Args:
    surface_form: surface form of a Turkish word that is to be morphologically
      analyzed.
    use_proper_feature: if true includes 'Proper' feature in the morphological
      analyses.

  Returns:
    Human-readable morphological analyses that the Turkish morphological
    analyzer yields for the given surface form. Returns an empty list if the
    given surface form is not accepted as a Turkish word form.
  """
  symbol_table = fst.ANALYZER.input_symbols()
  input_ = fst.compile(surface_form.encode("utf-8"), symbol_table)
  output = fst.compose(input_, fst.ANALYZER)

  if output.start() == -1:  # has no path to the accept state.
    return []

  human_readable = fst.extract_parses(
      output,
      output.start(),
      "olabel",
      symbol_table,
  )

  if not use_proper_feature:
    human_readable = (_remove_proper_feature(hr) for hr in human_readable)

  return sorted(set(human_readable))



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to generate Turkish word forms from analysis protobufs."""

import re
from typing import Generator

from turkish_morphology import analysis_pb2
from turkish_morphology import fst
from turkish_morphology import pretty_print

from external.openfst import pywrapfst

_Analysis = analysis_pb2.Analysis
_Ig = analysis_pb2.InflectionalGroup
_SymbolTable = pywrapfst.SymbolTable

_SYMBOLS_REGEX = re.compile(
    # First inflectional group.
    r"\(.+?\[[A-Z\.,:\(\)\'\-\"`\$]+?\]|"
    # Inflectional group boundaries.
    r"\)?\(\[[A-Z]+?\]|"
    # Derivational morphemes.
    r"-(?:[^\W\d_]|')+?\[[A-z]+?=[A-z]+?\]|"
    # Inflectional morphemes and features.
    r"\+(?:[^\W\d_]|['\.])*?\[[A-z]+?=[A-z0-9]+?\]|"
    # Proper noun analysis.
    r"\)\+\[Proper=(?:True|False)\]|"
    # Numbers.
    r"\d+(?:\[[A-Z]+?\])?|"
    # Parenthesis or decimal point separators.
    r"[\(\.,]")


def _lower(string: str) -> str:
  """Properly lowercase transforms Turkish string ("İ" -> "i", "I" -> "ı")."""
  return string.replace("İ", "i").replace("I", "ı").lower()


def _add_proper(analysis: _Analysis) -> None:
  """Adds the proper feature to the last inflectional group if it is missing."""
  last_ig = analysis.ig[-1]

  if last_ig.HasField("proper"):
    return analysis

  with_proper = _Analysis()
  with_proper.CopyFrom(analysis)
  with_proper.ig[-1].proper = last_ig.pos == "NNP"
  return with_proper


def _symbol_indices(analysis: _Analysis,
                    symbol_table: _SymbolTable) -> Generator[int, None, None]:
  """Generates the label indices for the symbols that construct the analysis."""
  human_readable = pretty_print.analysis(analysis)
  symbols = _SYMBOLS_REGEX.findall(human_readable)
  yield from map(symbol_table.find, symbols)


def surface_form(analysis: _Analysis) -> str:
  """Generates surface form for the given morphological analysis.

  This function assumes that input analysis protobuf is structurally
  well-formed, meaning that they should be first validated with
  //turkish_morphology:validate.py.

  Args:
    analysis: morphological analysis of a Turkish word from which surface form
      will be generated.

  Returns:
    Surface form of the Turkish word whose morphological analysis is the
    given morphological analysis. Returns an empty string if a surface form
    cannot be generated from the given morphological analysis.
  """
  symbol_table = fst.ANALYZER.input_symbols()
  symbol_indices = _symbol_indices(_add_proper(analysis), symbol_table)
  input_ = fst.compile(symbol_indices, symbol_table)
  output = fst.compose(fst.ANALYZER, input_)

  if output.start() == -1:  # has no path to the accept state.
    return ""

  surface_forms = fst.extract_parses(output, output.start(), "ilabel")
  surface_forms = list(set(map(_lower, surface_forms)))
  return surface_forms[0]



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to validate structural well-formedness of analysis protobufs."""

from typing import Optional

from turkish_morphology import analysis_pb2

_Affix = analysis_pb2.Affix
_Analysis = analysis_pb2.Analysis
_Feature = analysis_pb2.Feature
_Ig = analysis_pb2.InflectionalGroup
_Root = analysis_pb2.Root


class IllformedAnalysisError(Exception):
  """Raised when a human-readable analysis is structurally ill-formed."""


def _root(root: _Root) -> None:
  """Checks if root is structurally well-formed.

  Args:
    root: root analysis.

  Raises:
    IllformedAnalysisError: root is missing the morpheme field, or its
      morpheme field is set as empty string.
  """
  if not root.HasField("morpheme"):
    raise IllformedAnalysisError(f"Root is missing morpheme: '{root}'")

  if not root.morpheme:
    raise IllformedAnalysisError(f"Root morpheme is empty: '{root}'")


def _feature(feature: _Feature) -> None:
  """Checks if feature is structurally well-formed.

  Args:
    feature: morphological analysis feature.

  Raises:
    IllformedAnalysisError: feature is missing category or value field, or
      its category or value field is set as empty string.
  """
  if not feature.HasField("category"):
    raise IllformedAnalysisError(f"Feature is missing category: '{feature}'")

  if not feature.category:
    raise IllformedAnalysisError(f"Feature category is empty: '{feature}'")

  if not feature.HasField("value"):
    raise IllformedAnalysisError(f"Feature is missing value: '{feature}'")

  if not feature.value:
    raise IllformedAnalysisError(f"Feature value is empty: '{feature}'")


def _affix(affix: _Affix, derivational: Optional[bool] = False) -> None:
  """Checks if affix is structurally well-formed.

  Args:
    affix: affix analysis.
    derivational: if True, affix corresponds to a derivational feature.

  Raises:
    IllformedAnalysisError: affix is missing the feature field, or affix is
      derivational but its meta_morpheme field is missing or its meta_morpheme
      field is set as empty string.
  """
  if not affix.HasField("feature"):
    raise IllformedAnalysisError(f"Affix is missing feature: '{affix}'")

  _feature(affix.feature)

  if not derivational:
    return

  if not affix.HasField("meta_morpheme"):
    raise IllformedAnalysisError(
        f"Derivational affix is missing meta-morpheme: '{affix}'")

  if not affix.meta_morpheme:
    raise IllformedAnalysisError(
        f"Derivational affix meta-morpheme is empty: '{affix}'")


def _inflectional_group(ig: _Ig, position: int) -> None:
  """Checks if inflectional group is structurally well-formed.

  Args:
    ig: inflectional group analysis.
    position: index of the inflectional group w.r.t. the array of inflectional
      groups of the morphological analysis it belongs to.

  Raises:
    IllformedAnalysisError: inflectional group is missing part-of-speech tag,
      or its part-of-speech tag is set as empty string, or inflectional group
      is the first in morphological analysis and it is missing the root field,
      or inflectional group is derived and it is missing the derivation field.
  """
  if not ig.HasField("pos"):
    raise IllformedAnalysisError(
        f"Inflectional group {position + 1} is missing part-of-speech tag:"
        f" '{ig}'")

  if not ig.pos:
    raise IllformedAnalysisError(
        f"Inflectional group {position + 1} part-of-speech tag is empty:"
        f" '{ig}'")

  if position == 0:
    if not ig.HasField("root"):
      raise IllformedAnalysisError(
          f"Inflectional group {position + 1} is missing root: '{ig}'")

    _root(ig.root)
  else:
    if not ig.HasField("derivation"):
      raise IllformedAnalysisError(
          f"Inflectional group {position + 1} is missing derivational affix:"
          f" '{ig}'")

    _affix(ig.derivation, derivational=True)

  for inflection in ig.inflection:
    _affix(inflection)


def analysis(analysis: _Analysis) -> None:
  """Checks if analysis protobuf is structurally well-formed.

  Args:
    analysis: morphological analysis.

  Raises:
    IllformedAnalysisError: analysis is missing inflectional groups.
  """
  if not analysis.ig:
    raise IllformedAnalysisError(
        "Analysis is missing inflectional groups: 'analysis'")

  for position, ig in enumerate(analysis.ig):
    _inflectional_group(ig, position)



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

setuptools.setup(
    name="turkish-morphology",
    version="1.2.2",
    description="Turkish Morphology",
    long_description="A two-level morphological analyzer for Turkish.",
    url="https://github.com/google-research/turkish-morphology",
    download_url=("https://github.com/google-research/turkish-morphology/"
                  "releases"),
    license="Apache 2.0",
    packages=setuptools.find_packages(),
    package_data={
        "turkish_morphology": [
            "../external/openfst/pywrapfst.so",
            "../src/analyzer/bin/turkish.far",
        ],
    },
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Turkish",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=["absl-py", "protobuf"],
    python_requires='>=3.7',
)



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Turkish morphological analyzer FST utility functions."""

import os
import pathlib
from typing import Generator, Iterable, List, Optional

from external.openfst import pywrapfst

_Arc = pywrapfst.Arc
_Fst = pywrapfst.Fst
_SymbolTable = pywrapfst.SymbolTable

# Finite-state archive that contains the Turkish morphological analyzer FST
# is expected to be built into //src/analyzer/bin/turkish.far. See:
#     //src/analyzer/build.
_ROOT_DIR = pathlib.Path(__file__).parent.parent
_FAR_PATH = os.path.join(_ROOT_DIR, "src", "analyzer", "bin", "turkish.far")
_FST_NAME = "turkish_morphological_analyzer"


def _read_fst(far_path: str, fst_name: str) -> _Fst:
  """Reads FAR file from path and extracts the FST that has the sought name.

  Args:
    far_path: path to the finite-state archive that contains the sought FST.
    fst_name: name of the FST that will extracted from the finite-state archive.

  Returns:
    FST that has the sought name, which is extracted from the finite-state
    archive that is read from the path.
  """
  reader = pywrapfst.FarReader.open(far_path)
  return reader[fst_name]


ANALYZER = _read_fst(_FAR_PATH, _FST_NAME)


def compile(symbol_indices: Iterable[int], symbol_table: _SymbolTable) -> _Fst:
  """Compiles given sequnce of symbols in an FST.

  This function has similar behaviour to the StringCompiler that is initialized
  with BYTE StringTokenType, see:

      @openfst//src/include/fst/string.h

  Args:
    symbol_indices: indices of the sequence of symbols from the symbol table.
    symbol_table: symbol table which will be used to set as the symbol table
      of the compiled FST.

  Returns:
    FST that is compiled from the symbol indices.
  """
  fst = _Fst()
  last_state_index = fst.add_state()
  fst.set_start(last_state_index)

  for symbol_index in symbol_indices:
    arc = _Arc(symbol_index, symbol_index, 0, last_state_index + 1)
    fst.add_arc(last_state_index, arc)
    last_state_index = fst.add_state()

  fst.set_final(last_state_index, 0)
  fst.set_input_symbols(symbol_table)
  fst.set_output_symbols(symbol_table)
  return fst


def compose(this_fst: _Fst, that_fst: _Fst) -> _Fst:
  """Sorts the arcs of this FST and composes it with that FST."""
  this_fst.arcsort(sort_type="olabel")
  return pywrapfst.compose(this_fst, that_fst)


def extract_parses(
    fst: _Fst,
    state_index: int,
    label_type: str,
    symbol_table: Optional[_SymbolTable] = None,
    symbol_indices: Optional[List[int]] = []) -> Generator[str, None, None]:
  """Recursively extracts parses from the FST.

  This function extracts the parses by walking over all possible paths from the
  start state of the FST to its accept state. Joining the labels of the state
  transitions of these paths yield a parse.

  Args:
    fst: FST from which parses will be extracted.
    state_index: index of an FST state from which the paths to accept state will
      be traversed.
    label_type: from which tape parse symbols will be extracted (one of 'ilabel'
      or 'olabel').
    symbol_table: symbol table of the FST (only needed to look up complex
      symbols when label type is specified as 'olabel').
    symbol_indices: indices of symbols of a parse that are gathered from the
      start state of the FST till the state that has the given state index.

  Raises:
    AttributeError: label type is invalid.

  Yields:
    Parses that are extracted from the FST.
  """
  if label_type not in ("ilabel", "olabel"):
    raise AttributeError(f"Invalid label type: {label_type}")

  arcs = list(fst.arcs(state_index))

  if not arcs:  # is accept state, end of a parse.
    if label_type == "ilabel":
      yield bytes(symbol_indices).decode("utf-8")
    else:
      yield "".join(map(symbol_table.find, symbol_indices))

  for arc in arcs:
    new_symbol_indices = [s for s in symbol_indices]
    symbol_index = getattr(arc, label_type)

    if symbol_index != 0:  # skip <eps>.
      new_symbol_indices.append(symbol_index)

    yield from extract_parses(
        fst,
        arc.nextstate,
        label_type,
        symbol_table,
        new_symbol_indices,
    )



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to print human-readable analyses from analysis protobufs."""

from typing import Optional

from turkish_morphology import analysis_pb2

_Affix = analysis_pb2.Affix
_Analysis = analysis_pb2.Analysis
_Feature = analysis_pb2.Feature
_Ig = analysis_pb2.InflectionalGroup
_Root = analysis_pb2.Root


def _root(root: _Root) -> str:
  """Formats root analysis.

  Args:
    root: root analysis.

  Returns:
    Human-readable string that is used to represent the morpheme of the root
    analysis (e.g. "araba").
  """
  morpheme = root.morpheme
  return f"{morpheme}"


def _feature(feature: _Feature) -> str:
  """Formats morphological analysis feature.

  Args:
    feature: morphological analysis feature.

  Returns:
    Human-readable string that is used to represent the morphological analysis
    feature (e.g. '[Case=Abl]').
  """
  category = feature.category
  value = feature.value
  return f"[{category}={value}]"


def _affix(affix: _Affix, derivational: Optional[bool] = False) -> str:
  """Formats affix analysis.

  Args:
    affix: affix analysis.
    derivational: if True, affix corresponds to a derivational feature.

  Returns:
    Human-readable string that is used to represent the affix analysis (e.g.
    '+lAr[PersonNumber=A3pl]').
  """
  delimiter = "-" if derivational else "+"
  meta_morpheme = affix.meta_morpheme
  feature = _feature(affix.feature)
  return f"{delimiter}{meta_morpheme}{feature}"


def _inflectional_group(ig: _Ig, position: int) -> str:
  """Formats inflectional group analysis.

  Args:
    ig: inflectional group analysis.
    position: index of the inflectional group w.r.t. the array of inflectional
      groups of the morphological analysis it belongs to.

  Returns:
    Human-readable string that is used to represent the inflectional group
    analysis (e.g. '(araba[NN]+lAr[PersonNumber=A3pl]+[Possessive=Pnon]
    +DA[Case=Loc])').
  """
  pos = f"[{ig.pos}]"

  if position == 0:
    root = _root(ig.root)
    pos_root_derivation = f"{root}{pos}"
  else:
    derivation = _affix(ig.derivation, derivational=True)
    pos_root_derivation = f"{pos}{derivation}"

  inflections = "".join(_affix(i) for i in ig.inflection)

  if ig.HasField("proper"):
    proper = "+[Proper=True]" if ig.proper else "+[Proper=False]"
  else:
    proper = ""

  return f"({pos_root_derivation}{inflections}){proper}"


def analysis(analysis: _Analysis) -> str:
  """Constructs human-readable analysis from analysis protobuf.

  This function assumes that input analysis protobuf is structurally
  well-formed, meaning that they should be first validated with
  //turkish_morphology:validate.py.

  Args:
    analysis: morphological analysis.

  Returns:
    Human-readable string that is used to represent the morphological analysis
    (e.g. '(araba[NN]+lAr[PersonNumber=A3pl]+[Possessive=Pnon]+DA[Case=Loc])
    +[Proper=True]').
  """
  return "".join(_inflectional_group(ig, i) for i, ig in enumerate(analysis.ig))
